import logging
from typing import Any, List, Mapping, Optional, TypeVar, Union

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_nameToLevel = {
    'CRITICAL': CRITICAL,
    'FATAL': FATAL,
    'ERROR': ERROR,
    'WARN': WARNING,
    'WARNING': WARNING,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'NOTSET': NOTSET,
}


class MyLogger(object):
    def __init__(self, level: str = "INFO"):
        self.level = _nameToLevel.get(level)
        self.log_format = "%(asctime)s.%(msecs)03d [%(levelname)s] %(message)s"
        self.date_format = "%Y-%m-%d %H:%M:%S"
        logging.basicConfig(level=self.level,
                            format=self.log_format,
                            datefmt=self.date_format)
        self.logger = logging.getLogger(__name__)

    def info(self, msg: Optional[str] = None):
        self.logger.info(msg)

    def warning(self, msg: Optional[str] = None):
        self.logger.warning(msg)

    def error(self, msg: Optional[str] = None):
        self.logger.error(msg)

    def debug(self, msg: Optional[str] = None):
        self.logger.debug(msg)



import base64
import os
import time
from time import sleep
from typing import Optional

import docker

try:
    from IPython.display import display, HTML
    from IPython import get_ipython
except ImportError:
    print("ModuleNotFoundError: No module named 'IPython'")
import pandas as pd
import re
import requests
from tqdm import tqdm

from easy_kubeflow.utils import MyLogger

RepoTags = 'RepoTags'
Id = 'Id'
Created = 'Created'
Size = 'Size'
Names = 'Names'
Image = 'Image'
Command = 'Command'
Status = 'Status'
Ports = 'Ports'

NODE_HOST = '172.17.0.1'
Registry = 'harbor.stonewise.cn/'
Registry_Old = 'service.stonewise.cn:5000/'
Registry_Proxy = 'harbor-qzm.stonewise.cn/proxy_cache/'

USER_NAME = b'bGl1d2VpYmlu'
PASSWORD = b'U3RvbmV3aXNlQDIwMjExMA=='

namespace = os.environ["NB_PREFIX"].split("/")[-2]
notebook_name = os.environ["NB_PREFIX"].split("/")[-1]
pod_name = notebook_name + "-0"
my_container_name = notebook_name + "_" + pod_name + "_" + namespace

DEFAULT_DATA_CHUNK_SIZE = 1024 * 1024 * 100  # default chunk size 100MB
DEFAULT_TIMEOUT_SECONDS = 600  # default timeout for 10min

pd.set_option('display.max_columns', 100)
pd.set_option('display.max_rows', None)
pd.set_option('display.max_colwidth', 150)

_logger = MyLogger()


def _sub_str(character, start=0, lenght=None):
    try:
        if lenght:
            return character[start:start + lenght]
        else:
            return character[start:]
    except TypeError as te:
        print(te)
        return character


def _bit_convert_disk(bit_size):
    """
    convert bit to Mbit Gbit in disk
    base in disk is 1000 not 1024
    :param bit_size:
    :return:
    """
    try:
        if bit_size < 1000:
            return '{} B'.format(round(float(bit_size), 1))
        elif bit_size >= 1000 and bit_size < 1000000:
            return '{} KB'.format(round(float(bit_size / 1000), 1))
        elif bit_size >= 1000000 and bit_size < 1000000000:
            return '{} MB'.format(round(float(bit_size / 1000000), 1))
        else:
            return '{} GB'.format(round(float(bit_size / 1000000000), 1))
    except TypeError as te:
        print(te)
        return ''


def _str_convert_bit():
    pass


def _name_or_id(character):
    if _sub_str(character, start=0, lenght=7) == 'sha256:':
        return _sub_str(character, start=7, lenght=12)
    else:
        return character


def _select_iamge_info(dictionary):
    tag = dictionary.get(RepoTags)
    id = dictionary.get(Id)
    created = dictionary.get(Created)
    size = dictionary.get(Size)
    return {'REPOSITORY + TAG': tag, 'IMAGE ID': _sub_str(id, start=7, lenght=12),
            'CREATED': _sub_str(created, lenght=23), 'SIZE': _bit_convert_disk(bit_size=size)}


def _select_containers_info(dictionary):
    id = dictionary.get(Id)
    container_names = dictionary.get(Names)
    image_names = dictionary.get(Image)
    command = dictionary.get(Command)
    created = dictionary.get(Created)
    status = dictionary.get(Status)
    ports = dictionary.get(Ports)
    return {'CONTAINER ID': _sub_str(id, start=0, lenght=12),
            'CREATED': time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(created)), 'STATUS': status,
            'PORTS': '' if ports == [] else ports[0], 'IMAGE': _name_or_id(image_names), 'COMMAND': command,
            'CONTAINER NAME': container_names[0].strip('/')}


def _docker_progress(stream):
    stream_list = list(stream)
    _value = "initial"
    success = True
    for i in tqdm(range(len(stream_list))):
        value = list(stream_list[i].values())[0]
        if value.__class__.__name__ == 'str':
            value = value.strip('\n').strip('\r')
        if 'errorDetail' in stream_list[i].keys():
            success = False
            _logger.error(value)
        else:
            if value == _value:
                continue
            else:
                _value = value
                _logger.info(value)
        sleep(0.1)
    return success


def _del_contin_repeat_str(character, repeat):
    flag = False
    res = ''
    if character:
        for cha in character:
            if flag:
                if cha == repeat:
                    continue
                else:
                    res += cha
                    flag = False
            else:
                res += cha
                if cha == repeat:
                    flag = True
        return res
    else:
        return res


def _standard_dir(character):
    if character:
        if character[-1] != '/':
            _character = character + '/'
        else:
            _character = character
        return _del_contin_repeat_str(_character, '/')
    else:
        return ''


def _grep_match(character, string):
    matched = re.search(pattern=character, string=string, flags=0)
    if matched:
        return True
    else:
        return False


class EasyDocker(object):
    """Simple way to use docker in python"""

    def __init__(self, host: str = NODE_HOST, port: int = 8082, version: Optional[str] = "auto",
                 timeout: int = DEFAULT_TIMEOUT_SECONDS):
        self.host = host
        self.port = port
        self.base_url = "tcp://{host}:{port}".format(host=self.host, port=self.port)
        self.client = docker.DockerClient(base_url=self.base_url, version=version, timeout=timeout)
        self.client.login(username=base64.b64decode(USER_NAME).decode('ascii'),
                          password=base64.b64decode(PASSWORD).decode('ascii'),
                          registry='harbor.stonewise.cn')  # add authentication
        _logger.info("Connected to host docker successfully !")

    def close(self):
        """Closes all adapters and as such the session"""
        self.client.close()
        _logger.info("Closes all adapters and as such the session successfully !")

    def show_images(self, grep: Optional[str] = None):
        """
        Similar to cmd line ``docker images | grep xxx``.
        when name none, show all images
        :param grep: the item for grep. like `tensorflow/tensorflow:1.8.0` for simple match or
                     multiple repositories match like `tensorflow`, or multiple tags match like `1.8.0`...
        """
        images = self.client.images.list(name=None)
        res = [_select_iamge_info(_image.attrs) for _image in images]
        try:
            df = pd.DataFrame(res)
            df = df.explode("REPOSITORY + TAG").reset_index(drop=True)
            # filter NaN in dataframe
            df = df.dropna(axis=0, how='any')  # drop all rows that have any NaN values
            if grep:
                df = df[df["REPOSITORY + TAG"].apply(lambda x: _grep_match(character=grep, string=x))]
            if df.empty:
                _logger.warning("No matched repository or tag !")
        except KeyError:
            df = pd.DataFrame(columns=['REPOSITORY + TAG', 'IMAGE ID', 'CREATED', 'SIZE'])
            _logger.warning("No matched repository or tag !")
        
        try:
            use_notebook = type(get_ipython()).__module__.startswith('ipykernel.')
        except:
            use_notebook = False
        if use_notebook:
            display(HTML(df.to_html(index=False)))
        else:
            print(df.to_string(index=False))

    def pull_images(self, repository: Optional[str] = None, tag: Optional[str] = None):
        """
        pull images from cluster's own Registry: harbor.stonewise.cn/
        or harbor qzm proxy harbor-qzm.stonewise.cn/proxy_cache/
        Similar to cmd line ``docker pull``.
        :param repository: you'd better add Registry in your repository
        :param tag: Notice!!! if tag `None`, will push all tags repository from Registry
        """
        if repository:
            if _sub_str(repository, lenght=26) == Registry_Old or _sub_str(repository,
                                                                           lenght=20) == Registry or _sub_str(
                repository, lenght=36) == Registry_Proxy:
                _repository = repository
            else:
                _repository = Registry_Proxy + repository
                _logger.warning("Remember to add registry name in your repository, next time !")
            try:
                res = self.client.api.pull(repository=_repository, tag=tag, stream=True, decode=True)
                flag = _docker_progress(res)
                if flag:
                    _logger.info("Pull image successfully !")
                else:
                    _logger.error("Failed to pull image !")
            except requests.exceptions.HTTPError as e:
                print(e)
                _logger.error("Can't find repository in Registry !")
        else:
            _logger.error("You should at least input a repository !")

    def push_images(self, repository: Optional[str] = None, tag: Optional[str] = None):
        """
        push images to cluster's own Registry: harbor.stonewise.cn/
        Similar to cmd line ``docker push``.
        :param repository: you'd better add Registry in your repository
        :param tag: Notice!!! if tag `None`, will push all tags repository from Registry
        :return:
        """
        if repository:
            if _sub_str(repository, lenght=20) == Registry or _sub_str(repository, lenght=26) == Registry_Old:
                _repository = repository
            else:
                _repository = Registry + repository
                _logger.warning("Remember to add registry name in your repository, next time !")
            res = self.client.api.push(repository=_repository, tag=tag, stream=True, decode=True)
            flag = _docker_progress(res)
            if flag:
                _logger.info("Push image successfully !")
            else:
                _logger.error("Failed to push image !")
        else:
            _logger.error("You should at least input a repository !")

    def build_images(self, path: Optional[str] = None, dockerfile: Optional[str] = None,
                     repository: Optional[str] = None, tag: Optional[str] = None, rm: bool = True,
                     fileobj: Optional[object] = None, push_image: bool = False):
        """
        build remote repository by local config, dockerfile ...
        Similar to the ``docker build`` command.
        Either ``path`` or ``fileobj``
        needs to be set. ``path`` can be a local path (to a directory
        containing a Dockerfile) or a remote URL. ``fileobj`` must be a
        readable file-like object to a Dockerfile.

        :param path: Path to the directory containing the Dockerfile
        :param dockerfile: name of docker file.
        :param repository: the repository to build
        :param tag: the tag to tag
        :param rm: Remove intermediate containers. The ``docker build``
                   command now defaults to ``--rm=true``, but we have kept the old
                   default of `False` to preserve backward compatibility
        :param fileobj: A file object to use as the Dockerfile. (Or a file-like
                        object)
        :param push_image: push repository to Registry, default false.
        """
        if repository:
            if _sub_str(repository, lenght=20) == Registry or _sub_str(repository, lenght=26) == Registry_Old:
                _repository = repository
            else:
                _repository = Registry + repository
                _logger.warning("Remember to add registry name in your repository, next time !")
            if tag:
                _tag = tag
            else:
                _tag = 'latest'
                _logger.warning("add latest as default tag !")
            streamer = self.client.api.build(path=path, dockerfile=dockerfile, tag=_repository + ':' + _tag, rm=rm,
                                             fileobj=fileobj, decode=True)
            flag = _docker_progress(streamer)
            if flag:
                _logger.info("Build image successfully !")
            else:
                _logger.error("Failed to build image !")
            if push_image:
                self.push_images(repository=repository, tag=tag)  # push built image, just now to Registry
        else:
            _logger.error("You should at least input a repository !")

    def tag_images(self, original_repository: Optional[str] = None, original_tag: Optional[str] = None,
                   target_repository: Optional[str] = None, target_tag: Optional[str] = None, force: bool = False):
        """
        Tag an image into a repository. Similar to the ``docker tag`` command.
        if target_repository target_tag both `None`, default tagged Registry + original_repository:original_tag
        :param original_repository:
        :param original_tag:
        :param target_repository:
        :param target_tag:
        :param force: force tag, default false
        """
        if target_repository:
            _target_repository = target_repository
            _target_tag = target_tag
        else:
            try:
                if _sub_str(original_repository, lenght=20) == Registry or _sub_str(original_repository,
                                                                                    lenght=36) == Registry_Proxy:
                    _target_repository = original_repository
                else:
                    _target_repository = Registry + original_repository
                _logger.warning("Use original repository as default target repository !")
            except TypeError:
                _target_repository = None
            if target_tag:
                _target_tag = target_tag
            else:
                _target_tag = original_tag

        if original_repository:
            if original_tag:
                _original = original_repository + ':' + original_tag
            else:
                _original = original_repository
            self.client.api.tag(image=_original, repository=_target_repository, tag=_target_tag, force=force)
            _logger.info("Tag repository successfully !")
        else:
            _logger.error("You should have an original repository !")

    def remove_images(self, repository: Optional[str] = None, tag: Optional[str] = None, image_id: Optional[str] = None,
                      force: bool = False):
        """
        Two options to remove image: use `repository:tag` or `image id`.
        Similar to the ``docker rmi `` command.
        :param repository:
        :param tag:
        :param image_id:
        :param force: remove image by force, default false
        """
        if repository:
            if tag:
                _image = repository + ':' + tag
            else:
                _image = repository
        else:
            if image_id:
                _image = image_id
            else:
                _image = None
        if _image:
            try:
                self.client.api.remove_image(image=_image, force=force)
                _logger.info("Remove image {image} successfully !".format(image=_image))
            except requests.exceptions.HTTPError as ex:
                _logger.error(ex)
        else:
            _logger.error("You should at least input a repository or image id!")

    def save_images(self, repository: Optional[str] = None, tag: Optional[str] = None,
                    chunk_size: Optional[int] = DEFAULT_DATA_CHUNK_SIZE, save_dir: Optional[str] = None,
                    file_name: Optional[str] = None):
        """
        Save an image as tar file in container's pv. Similar to the ``docker save`` command.
        :param repository:
        :param tag:
        :param chunk_size: The number of bytes returned by each iteration
                           of the generator. If ``None``, data will be streamed as it is
                           received. Default: 100 MB
        :param save_dir: if None, the dir script exist
        :param file_name: xxx.tar or xxx.tar.gz, default name `default.tar.gz`
        :return:
        """
        if repository:
            _repository = repository
            if tag:
                _tag = tag
            else:
                _tag = 'latest'
                _logger.warning("add `latest` as default tag !")
            _image = _repository + ":" + _tag

            if file_name:
                _file_name = file_name
            else:
                _file_name = 'default.tar.gz'
                _logger.warning(
                    "Use default file name `default.tar.gz` remember to add tar file name for saving image, next time !")
            if save_dir:
                if os.path.isdir(save_dir):
                    _save_dir = _standard_dir(save_dir)
                else:
                    _save_dir = ''
            else:
                _save_dir = ''  # use current dir
            _save_file = _save_dir + _file_name
            raw_data = self.client.api.get_image(image=_image, chunk_size=chunk_size)
            try:
                with open(_save_file, 'wb') as image_tar:
                    for chunk in raw_data:
                        image_tar.write(chunk)
                _logger.info("Saved image tar file {file} to dir: {dir} successfully !".format(file=_file_name,
                                                                                               dir=_save_dir if _save_dir else 'current dir'))
            except requests.exceptions.HTTPError:
                _logger.error("No such image {repository}:{tag} !".format(repository=_repository, tag=_tag))
        else:
            _logger.error("You should at least input a repository !")

    def load_images(self, file_name: Optional[str] = None, file_dir: str = './'):
        """
        Similar to ``docker load``.
        :param file_name:
        :param file_dir: you'd better user absolute dir, default './'
        """
        path = file_dir + file_name
        with open(path, 'br') as binary_file:
            res = self.client.api.load_image(binary_file)
        flag = _docker_progress(res)
        if flag:
            _logger.info("Load image successfully !")
        else:
            _logger.error("Failed to load image !")

    def show_containers(self, all: bool = False, limit: int = -1,
                        filters: Optional[dict] = {"name": my_container_name}):
        """
        List containers. Similar to the ``docker ps`` command.
        :param all: Show all containers. Only running containers are shown
                    by default
        :param limit: Show `limit` last created containers, include
                      non-running ones
        :param filters: Filters to be processed on the image list.
                Available filters:

                - `exited` (int): Only containers with specified exit code
                - `status` (str): One of ``restarting``, ``running``,
                    ``paused``, ``exited``
                - `label` (str|list): format either ``"key"``, ``"key=value"``
                    or a list of such.
                - `id` (str): The id of the container.
                - `name` (str): The name of the container.
                - `ancestor` (str): Filter by container ancestor. Format of
                    ``<image-name>[:tag]``, ``<image-id>``, or
                    ``<image@digest>``.
                - `before` (str): Only containers created before a particular
                    container. Give the container name or id.
                - `since` (str): Only containers created after a particular
                    container. Give container name or id.

                A comprehensive list can be found in the documentation for
                `docker ps
                <https://docs.docker.com/engine/reference/commandline/ps>`_.
        """
        try:
            containers = self.client.api.containers(all=all, limit=limit, filters=filters)
            res = [_select_containers_info(_container) for _container in containers]
        except requests.exceptions.HTTPError:
            _logger.error("Wrong use of params ！")
            res = []
        if res == []:
            df = pd.DataFrame(
                columns=['CONTAINER ID', 'CREATED', 'STATUS', 'PORTS', 'IMAGE', 'COMMAND', 'CONTAINER NAME'])
            _logger.warning("Can't find select containers !")
        else:
            df = pd.DataFrame(res)
        
        try:
            use_notebook = type(get_ipython()).__module__.startswith('ipykernel.')
        except:
            use_notebook = False
            
        if use_notebook:
            display(HTML(df.to_html(index=False)))
        else:
            pd.set_option("expand_frame_repr", False)
            pd.set_option("max_colwidth", 22)
            pd.set_option("display.colheader_justify", 'center')
            # to remove index
            blankIndex=[''] * len(df)
            df.index=blankIndex
            print(df)
            pd.reset_option('expand_frame_repr')
            pd.reset_option('max_colwidth')
            pd.reset_option('display.colheader_justify')

    def commit_containers(self, container: Optional[str] = Names, repository: Optional[str] = None,
                          tag: Optional[str] = None, push_image: bool = False):
        """
        Commit a container to an image. Similar to the ``docker commit`` command.
        :param container: container id
        :param repository: The repository to push the image to (you'd better add Registry in your repository)
        :param tag: The tag to push
        ":param push_image: push repository to Registry, default false.
        """
        if repository:
            if _sub_str(repository, lenght=20) == Registry or _sub_str(repository, lenght=26) == Registry_Old:
                _repository = repository
            else:
                _repository = Registry + repository
                _logger.warning("Remember to add registry name in your repository, next time !")
            if tag:
                _tag = tag
            else:
                _tag = 'latest'
                _logger.warning("add latest as default tag !")
            try:
                self.client.api.commit(container=container, repository=_repository, tag=_tag)
                _logger.info("Commit container successfully !")
                _logger.info("Committed image:   {repository}:{tag}".format(repository=_repository,
                                                                            tag=_tag
                                                                            )
                             )
                if push_image:
                    self.push_images(repository=_repository, tag=_tag)  # push commited image, just now to Registry
            except requests.exceptions.HTTPError as e:
                _logger.error(e)
                # _logger.error("No such container: {container}".format(container=container))
        else:
            _logger.error("You should at least input a repository !")



import inspect
from typing import Callable, Optional
from uuid import uuid4

import kfp
from kfp import compiler
import kfp.dsl as dsl
import kfp.notebook
from kfp_server_api.exceptions import ApiException
from kubernetes import client as k8s_client

from easy_kubeflow.utils import MyLogger

Strategies = {
    '1080': {
        'gpu-type': '1080'
    },
    '2080': {
        'gpu-type': '2080'
    },
    'v100s': {
        'gpu-type': 'v100s'
    },
    'unshared_node': {
        'func': 'job'
    }
}

# select default resource
CPU_REQUEST = '0.5'
CPU_LIMIT = '1'
MEM_REQUEST = '1Gi'
MEM_LMIT = '2Gi'

# set nfs
NFS_MASTER_HOST = '10.55.22.78'
NFS_PATH = '/kubeflow-pv-pro/'

_logger = MyLogger()


def _valid_strategy(strategy):
    if strategy in Strategies:
        return True
    else:
        return False


def _get_strategy_key_value(dic):
    for key, value in dic.items():
        return key, value


def _input_arg_path(arg):
    """
    ping to args path
    :param arg:
    :return: ars path
    """
    return dsl.InputArgumentPath(argument=arg)


def _is_PipelineParam(obj):
    if obj:
        if obj.__class__.__name__ == 'PipelineParam':
            return True
        else:
            return False
    else:
        _logger.error("Invalid args")
        return False


def _dict2list(dic: dict = None):
    ls = list()
    if dic:
        for key, value in dic.items():
            ls.append(key)
            ls.append(value)
        return ls
    else:
        return None


def _uniform_data_path(dic: dict = None):
    re_dic = dict()
    if dic:
        for key, value in dic.items():
            re_dic[key] = value
        return re_dic
    else:
        return None


def _is_url(character):
    if character.startswith('http'):
        return True
    else:
        return False


_pipeline_decorator_handler = None


class EasyPipelines(object):
    """easy way to kubeflow pipelines"""

    def __init__(self, description: Optional[str] = None):
        """
        :param description: description for your experiment, default no description.
        """
        self.client = None
        self.experiment_name = 'Default'
        self.description = description
        self.components_list = list()

    @property
    def experiment(self):
        return self.experiment_name

    @experiment.setter
    def experiment(self, name: Optional[str] = None):
        """
        set experiment
        :param name: `Default` is default experiment name
        """
        self.experiment_name = name
        if not self.client:
            self.client = kfp.Client()
        self.experiment_obj = self.client.create_experiment(name=self.experiment_name, description=self.description)
        _logger.info("Connected to kubeflow pipelines client successfully !")

    @property
    def components(self):
        return self.components_list

    @components.setter
    def components(self, comp):
        self.components_list.append(comp)

    def dec_func(self):
        return dsl.pipeline()

    def input_arg_path(self, arg):
        """
        ping to args path
        :param arg:
        :return: ars path
        """
        return dsl.InputArgumentPath(argument=arg)

    def pipeline(self,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 output_directory: Optional[str] = None):
        """Decorator of pipeline functions.

        Example
          ::

            @pipeline(
              name='my awesome pipeline',
              description='Is it really awesome?'
              output_directory='gs://my-bucket/my-output-path'
            )
            def my_pipeline(a: PipelineParam, b: PipelineParam):
              ...

        Args:
          name: The pipeline name. Default to a sanitized version of the function
            name.
          description: Optionally, a human-readable description of the pipeline.
          output_directory: The root directory to generate input/output URI under this
            pipeline. This is required if input/output URI placeholder is used in this
            pipeline.
        """

        def _pipeline(func: Callable):
            if name:
                func._component_human_name = name
            if description:
                func._component_description = description
            if output_directory:
                func.output_directory = output_directory

            if _pipeline_decorator_handler:
                return _pipeline_decorator_handler(func) or func
            else:
                return func

        return _pipeline

    def compiler_pipeline(self, pipeline_func, file_name):
        """
        compile pipeline in notebook
        :param pipeline_func: Pipeline functions with @pipeline decorator.
        :param file_name: The output workflow tar.gz file path. for example, "~/a.tar.gz"
        """
        compiler.Compiler().compile(pipeline_func=pipeline_func, package_path=file_name)
        pass

    def run_pipeline(self, run_name: str = None, pipeline_package: str = None):
        """
        run pipeline in notebook
        :param run_name: run name for your experiment, runs will be nested in the same experiment.
        :param pipeline_package: Local path of the pipeline package(
                                 the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
        """
        if not self.client:
            self.client = kfp.Client()
        try:
            self.client.run_pipeline(experiment_id=self.experiment_obj.id, job_name=run_name,
                                     pipeline_package_path=pipeline_package)
        except ApiException as ae:
            _logger.error(ae)

    def _pars_comp(self, comp_list: list = None):
        if comp_list:
            for comp in comp_list:
                print(comp.name)
        pass

    def _dag(self):
        pass

    def load_component(self, component_file: str = None):
        """
        load reuse component from file or url
        :param component_file: url or local dir
        :return: reuse component factory func
        """
        func = None
        if component_file:
            is_url = _is_url(component_file)
            if is_url:
                func = kfp.components.load_component_from_url(url=component_file)
            else:
                func = kfp.components.load_component_from_file(filename=component_file)
            _logger.info(
                "args for your component factory func are : `%s`" % inspect.getfullargspec(func=func).annotations)
            _logger.info("You can use class `ReuseComponent` to build pipelines component !")
        else:
            _logger.error("at least, you need to add a component file !")
        return func


class Component(object):
    """definition for pipelines component"""

    def __init__(self, name: str = None, image: str = None,
                 command: Optional[list] = None,
                 arguments: Optional[dict] = None,
                 file_outputs: Optional[dict] = None,
                 nfs_path: str = NFS_PATH,
                 nfs_master_host: str = NFS_MASTER_HOST):
        """
        Component object
        :param name: name of component
        :param image: image for component container
        :param command: [list] args to run component container, default None.
                        the command to run in the container. If None, uses default CMD
                        in defined in container.
        :param arguments: [dict] args to run component container, default None
        :param file_outputs: [dict] out put file of component container, default None
        :param nfs_path: default root path defined by cluster admin
        :param nfs_master_host: default host selected by cluster admin
        """
        self.name = name
        self.uuid = str(uuid4())
        self.image = image
        self.command = command
        self.arguments = arguments
        self.file_outputs = file_outputs
        self.nfs_path = nfs_path
        self.nfs_master_host = nfs_master_host
        self.op = dsl.ContainerOp(
            name=self.name,
            image=self.image,
            command=self.command,
            arguments=_dict2list(self.arguments),
            file_outputs=self.file_outputs
        )

    def default_op(self) -> 'dsl.ContainerOp':
        """
        default pipelines component containerOp
        :return: component containerOp
        """
        _logger.info("Component op of `%s` init ..." % self.name)
        _logger.info("Container mount path is: %s" % self.nfs_path)
        self._pull_image_policy()
        self._add_volume()
        return self.op

    def udf_op(self, request_cpu: Optional[str] = None, request_mem: Optional[str] = None,
               request_gpu: Optional[str] = None,
               strategies: Optional[list] = None) -> 'dsl.ContainerOp':
        """
        user define pipelines component containerOp
        :param request_cpu: a string can be a number followed by "m", like `300m`, or `0.3` which equals `300m`
        :param request_mem: a string which can be a number or a number followed by one of "E", "P", "T", "G", "M", "K".
                            like `1Gi`, `1Mi`, `1Ki`
        :param request_gpu: str(number) like '1', '4', '8' means the amount of gpu you use
        :param strategies: node select strategy, defined by cluster admin. strategy: `"big_data"` is available.
        :return: component containerOp
        """
        self.default_op()
        self._node_selecter(strategies=strategies)
        self._resource_limit(request_cpu=request_cpu, request_mem=request_mem, request_gpu=request_gpu)
        return self.op

    def after(self, ops_list: Optional[list] = None, *ops):
        """
        add ops list like parallel ops
        :param ops_list: ops list
        :param ops: ops
        :return:
        """
        if ops_list:
            for item in ops_list:
                self.op.dependent_names.append(item.name)
            for item in ops:
                self.op.dependent_names.append(item.name)
        return self.op

    def _pull_image_policy(self, policy: str = 'Always'):
        """
        default image pull policy: `Always`
        """
        self.op.container.image_pull_policy = policy
        _logger.info("Use image pull policy: %s" % policy)

    def _add_volume(self):
        """add nfs volume, volume and mount volume has the same path"""
        self.op.add_volume(
            k8s_client.V1Volume(name=self.uuid + '-pv',
                                nfs=k8s_client.V1NFSVolumeSource(
                                    path=self.nfs_path,
                                    server=self.nfs_master_host
                                )
                                )
        ) \
            .container \
            .add_volume_mount(
            k8s_client.V1VolumeMount(
                mount_path=self.nfs_path,
                name=self.uuid + '-pv')
        )

        # add shm for pytorch multiple worker use
        self.op.add_volume(
            k8s_client.V1Volume(name='dshm',
                                empty_dir=k8s_client.V1EmptyDirVolumeSource(
                                    medium='Memory'
                                )
                                )
        ) \
            .container \
            .add_volume_mount(
            k8s_client.V1VolumeMount(
                mount_path='/dev/shm',
                name='dshm'
            )
        )
        _logger.info("Nfs server master host is: %s" % self.nfs_master_host)
        _logger.info("Nfs server mount path: %s" % self.nfs_path)

    def _node_selecter(self, strategies: Optional[list] = None):
        """
        add pod's node select strategies
        :param strategies: strategy list
        """
        if strategies:
            for strategy in strategies:
                if _valid_strategy(strategy):
                    key, value = _get_strategy_key_value(Strategies.get(strategy))
                    self.op.add_node_selector_constraint(key, value)
                    _logger.info("Add strategy: %s" % strategy)
                else:
                    _logger.error("strategy: %s not found in Strategies" % strategy)
        else:
            _logger.warning("Use default strategy")

    def _resource_limit(self, request_cpu: Optional[str] = None,
                        request_mem: Optional[str] = None,
                        request_gpu: Optional[str] = None):
        """
        resource: cpu, memory
        when user set resource, we recommend `request` equals to `limit`
        :param request_cpu: a string can be a number followed by "m", like `300m`, or `0.3` which equals `300m`
        :param request_mem: a string which can be an integer number or an integer number followed by one of "E",
                            "P", "T", "G", "M", "K". like `1Gi`, `1Mi`, `1Ki`
        :param request_gpu: str(number) like '1', '4', '8' means the amount of gpu you use
        """
        # config for cpu
        if request_cpu:
            self.op.container.set_cpu_request(
                request_cpu) \
                .set_cpu_limit(
                request_cpu)
            _logger.info("Set cpu request: %s" % request_cpu)
            _logger.info("Set cpu limit: %s" % request_cpu)
        else:
            self.op.container.set_cpu_request(
                CPU_REQUEST) \
                .set_cpu_limit(
                CPU_LIMIT)
            _logger.warning("Use default cpu request: %s" % CPU_REQUEST)
            _logger.warning("Use default cpu limit: %s" % CPU_LIMIT)
        # config for memory
        if request_mem:
            self.op.container.set_memory_request(
                request_mem) \
                .set_memory_limit(
                request_mem)
            _logger.info("Set mem request: %s" % request_mem)
            _logger.info("Set mem limit: %s" % request_mem)
        else:
            self.op.container.set_memory_request(
                MEM_REQUEST) \
                .set_memory_limit(
                MEM_LMIT)
            _logger.warning("Use default mem request: %s" % MEM_REQUEST)
            _logger.warning("Use default mem limit: %s" % MEM_LMIT)
        # config for gpu
        if request_gpu:
            self.op.container.set_gpu_limit(
                gpu=request_gpu)
            _logger.info("Set gpu request amount: %s" % request_gpu)
            _logger.info("Set gpu limit amount: %s" % request_gpu)
        else:
            _logger.warning("Use default gpu request: 0")
            _logger.warning("Use default gpu limit: 0")


# TODO: ReuseComponent use command
class ReuseComponent(object):
    """definition for reuse pipelines component"""

    def __init__(self, component_factory_func: Callable = None,
                 arguments: Optional[dict] = None,
                 command: Optional[list] = None,
                 nfs_path: str = NFS_PATH,
                 nfs_master_host: str = NFS_MASTER_HOST):
        """
        Reuse component object
        :param component_factory_func: factory func to build reuse component op
        :param command: [list] args to run component container, default None.
                        the command to run in the container. If None, uses default CMD
                        in defined in container.
        :param arguments: [dict] args to run reuse component container, default None
        :param nfs_path: default root path defined by cluster admin
        :param nfs_master_host: default host selected by cluster admin
        """
        self.uuid = str(uuid4())
        self.arguments = arguments
        self.command = command
        self.factory_func = component_factory_func
        self.nfs_path = nfs_path
        self.nfs_master_host = nfs_master_host
        self.op = self.factory_func(**_uniform_data_path(self.arguments))

    @staticmethod
    def _uniform_name(name: str = None):
        if name:
            name = name.lower()
            return '-'.join(name.split())
        else:
            return name

    def default_op(self) -> 'dsl.ContainerOp':
        """
        default pipelines component containerOp
        :return: component containerOp
        """
        _logger.info("ReuseComponent op of `%s` init ..." % self._uniform_name(self.factory_func.__name__))
        _logger.info("Container mount path is: %s" % self.nfs_path)
        self._pull_image_policy()
        self._add_volume()
        self._add_command(self.command)
        return self.op

    def udf_op(self, request_cpu: Optional[str] = None, request_mem: Optional[str] = None,
               request_gpu: Optional[str] = None,
               strategies: Optional[list] = None) -> 'dsl.ContainerOp':
        """
        user define pipelines component containerOp
        :param request_cpu: a string can be a number followed by "m", like `300m`, or `0.3` which equals `300m`
        :param request_mem: a string which can be a number or a number followed by one of "E", "P", "T", "G", "M", "K".
                            like `1Gi`, `1Mi`, `1Ki`
        :param request_gpu: str(number) like '1', '4', '8' means the amount of gpu you use
        :param strategies: node select strategy, defined by cluster admin. strategy: `"big_data"` is available.
        :return: component containerOp
        """
        self.default_op()
        self._node_selecter(strategies=strategies)
        self._resource_limit(request_cpu=request_cpu, request_mem=request_mem, request_gpu=request_gpu)
        return self.op

    def after(self, ops_list: Optional[list] = None, *ops):
        """
        add ops list like parallel ops
        :param ops_list: ops list
        :param ops: ops
        :return:
        """
        if ops_list:
            for item in ops_list:
                self.op.dependent_names.append(item.name)
            for item in ops:
                self.op.dependent_names.append(item.name)
        return self.op

    def _pull_image_policy(self, policy: str = 'Always'):
        """
        default image pull policy: `Always`
        """
        self.op.container.image_pull_policy = policy
        _logger.info("Use image pull policy: %s" % policy)

    def _add_command(self, command: list = None):
        """
        default command: None
        """
        if command:
            self.op.container.command = command
            _logger.info("Use command: %s" % command)
        else:
            _logger.warning("Use default command")

    def _add_volume(self):
        """add nfs volume, volume and mount volume has the same path"""
        self.op.add_volume(
            k8s_client.V1Volume(name=self.uuid + '-pv',
                                nfs=k8s_client.V1NFSVolumeSource(
                                    path=self.nfs_path,
                                    server=self.nfs_master_host
                                )
                                )
        ) \
            .container \
            .add_volume_mount(
            k8s_client.V1VolumeMount(
                mount_path=self.nfs_path,
                name=self.uuid + '-pv')
        )

        # add shm for pytorch multiple worker use
        self.op.add_volume(
            k8s_client.V1Volume(name='dshm',
                                empty_dir=k8s_client.V1EmptyDirVolumeSource(
                                    medium='Memory'
                                )
                                )
        ) \
            .container \
            .add_volume_mount(
            k8s_client.V1VolumeMount(
                mount_path='/dev/shm',
                name='dshm'
            )
        )
        _logger.info("Nfs server master host is: %s" % self.nfs_master_host)
        _logger.info("Nfs server mount path: %s" % self.nfs_path)

    def _node_selecter(self, strategies: Optional[list] = None):
        """
        add pod's node select strategies
        :param strategies: strategy list
        """
        if strategies:
            for strategy in strategies:
                if _valid_strategy(strategy):
                    key, value = _get_strategy_key_value(Strategies.get(strategy))
                    self.op.add_node_selector_constraint(key, value)
                    _logger.info("Add strategy: %s" % strategy)
                else:
                    _logger.error("strategy: %s not found in Strategies" % strategy)
        else:
            _logger.warning("Use default strategy")

    def _resource_limit(self, request_cpu: Optional[str] = None,
                        request_mem: Optional[str] = None,
                        request_gpu: Optional[str] = None):
        """
        resource: cpu, memory
        when user set resource, we recommend `request` equals to `limit`
        :param request_cpu: a string can be a number followed by "m", like `300m`, or `0.3` which equals `300m`
        :param request_mem: a string which can be an integer number or an integer number followed by one of "E",
                            "P", "T", "G", "M", "K". like `1Gi`, `1Mi`, `1Ki`
        :param request_gpu: str(number) like '1', '4', '8' means the amount of gpu you use
        """
        # config for cpu
        if request_cpu:
            self.op.container.set_cpu_request(
                request_cpu) \
                .set_cpu_limit(
                request_cpu)
            _logger.info("Set cpu request: %s" % request_cpu)
            _logger.info("Set cpu limit: %s" % request_cpu)
        else:
            self.op.container.set_cpu_request(
                CPU_REQUEST) \
                .set_cpu_limit(
                CPU_LIMIT)
            _logger.warning("Use default cpu request: %s" % CPU_REQUEST)
            _logger.warning("Use default cpu limit: %s" % CPU_LIMIT)
        # config for memory
        if request_mem:
            self.op.container.set_memory_request(
                request_mem) \
                .set_memory_limit(
                request_mem)
            _logger.info("Set mem request: %s" % request_mem)
            _logger.info("Set mem limit: %s" % request_mem)
        else:
            self.op.container.set_memory_request(
                MEM_REQUEST) \
                .set_memory_limit(
                MEM_LMIT)
            _logger.warning("Use default mem request: %s" % MEM_REQUEST)
            _logger.warning("Use default mem limit: %s" % MEM_LMIT)
        # config for gpu
        if request_gpu:
            self.op.container.set_gpu_limit(
                gpu=request_gpu)
            _logger.info("Set gpu request amount: %s" % request_gpu)
            _logger.info("Set gpu limit amount: %s" % request_gpu)
        else:
            _logger.warning("Use default gpu request: 0")
            _logger.warning("Use default gpu limit: 0")



import setuptools

with open("README.md", "r") as fh:
    long_description = None

setuptools.setup(
    name="easy-kubeflow",  # Replace with your own name
    version="0.0.9",
    author="CrazyBean",
    author_email="liuweibin@stonewise.cn",
    description="sdk help users for a better use of kubeflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://service.stonewise.cn:8029/liuweibin/easy-kubeflow.git",
    install_requires=[
        'docker>=4.2.1',
        'kfp==1.3.0',
        'pandas>=1.1.5',
        'tqdm>=4.56.0'
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)
